{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Link to this document's Jupyter Notebook](./0310-CUDA_Tiling_in-class-assignment.ipynb)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In order to successfully complete this assignment you need to participate both individually and in groups during class.   If you attend class in-person then have one of the instructors check your notebook and sign you out before leaving class on **Wednesday March 10**. If you are attending asynchronously, turn in your assignment using D2L no later than **_11:59pm on Wednesday March 10**. "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "---\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# In-Class Assignment: CUDA Memory and Tiling\n",
    "\n",
    "<img alt=\"picture of a tile mosaic designed to motivate this topic.\" src=\"https://www.appianimosaic.com/uploads/2016-6-24/1920-300/per_downolad_2n.jpg\">\n",
    "\n",
    "\n",
    "Image from: https://www.appianimosaic.com/\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Agenda for today's class (70 minutes)\n",
    "\n",
    "\n",
    "1. (20 minutes) [Pre class Review](#Pre-class-Review)\n",
    "2. (20 minutes) [Tile example](#Tile-example)\n",
    "3. (30 minutes) [1D wave Cuda Code Optimization](#1D-wave-Cuda-Code-Optimization)\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "---\n",
    "<a name=Pre-class-Review></a>\n",
    "# 1. Pre class Review\n",
    "\n",
    "\n",
    "\n",
    "- [00309--CUDA_Tiling_pre-class-assignment](0309--CUDA_Tiling_pre-class-assignment.ipynb)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "\n",
    "\n",
    "---\n",
    "<a name=Tile-example></a>\n",
    "# 2. Tile example\n",
    "\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%writefile NCode/tiled_transpose.cu\n",
    "\n",
    "#include <iostream>\n",
    "#include <cuda.h>\n",
    "#include <chrono>\n",
    "#define CUDA_CALL(x) {cudaError_t cuda_error__ = (x); if (cuda_error__) { fprintf(stderr, \"CUDA error: \" #x \" returned \\\"%s\\\"\\n\", cudaGetErrorString(cuda_error__)); fflush(stderr); exit(cuda_error__); } }\n",
    "\n",
    "using namespace std;\n",
    "\n",
    "const int BLOCKDIM = 32; \n",
    "\n",
    "__global__ void transpose(const double *in_d, double * out_d, int row, int col)\n",
    "{\n",
    "   int x = blockIdx.x * blockDim.x + threadIdx.x;\n",
    "   int y = blockIdx.y * blockDim.y + threadIdx.y;\n",
    "   if (x < col && y < row) \n",
    "       out_d[y+col*x] = in_d[x+row*y];\n",
    "}\n",
    "\n",
    "__global__ void tiled_transpose(const double *in_d, double * out_d, int row, int col)\n",
    "{\n",
    "   int x = blockIdx.x * BLOCKDIM + threadIdx.x;\n",
    "   int y = blockIdx.y * BLOCKDIM + threadIdx.y;\n",
    "    \n",
    "   int x2 = blockIdx.y * BLOCKDIM + threadIdx.x;\n",
    "   int y2 = blockIdx.x * BLOCKDIM + threadIdx.y;\n",
    "    \n",
    "   __shared__ double in_local[BLOCKDIM][BLOCKDIM];\n",
    "   __shared__ double out_local[BLOCKDIM][BLOCKDIM];\n",
    "\n",
    "   if (x < col && y < row) {\n",
    "       in_local[threadIdx.x][threadIdx.y] = in_d[x+row*y];\n",
    "       __syncthreads();\n",
    "\n",
    "       out_local[threadIdx.y][threadIdx.x] = in_local[threadIdx.x][threadIdx.y];\n",
    "       __syncthreads();\n",
    "\n",
    "       out_d[x2+col*y2] = out_local[threadIdx.x][threadIdx.y];\n",
    "   }\n",
    "}\n",
    "\n",
    "__global__ void transpose_symmetric(double *in_d, double * out_d, int row, int col)\n",
    "{\n",
    "   int x = blockIdx.x * blockDim.x + threadIdx.x;\n",
    "   int y = blockIdx.y * blockDim.y + threadIdx.y;\n",
    "   if (x < col && y < row) {\n",
    "       if (x < y) { \n",
    "           double temp =  in_d[y+col*x];\n",
    "           in_d[y+col*x] = in_d[x+row*y];\n",
    "           in_d[x+row*y] = temp;\n",
    "       }\n",
    "   }\n",
    "}\n",
    "\n",
    "\n",
    "\n",
    "int main(int argc,char **argv)\n",
    "{\n",
    "    std::cout << \"Begin\\n\";\n",
    "   int sz_x=BLOCKDIM*300;\n",
    "   int sz_y=BLOCKDIM*300;\n",
    "   int nBytes = sz_x*sz_y*sizeof(double);\n",
    "   int block_size = BLOCKDIM;\n",
    "   double *m_h = (double *)malloc(nBytes);\n",
    "   double * in_d;\n",
    "   double * out_d;\n",
    "   int count = 0;\n",
    "   for (int i=0; i < sz_x*sz_y; i++){\n",
    "       m_h[i] = count;\n",
    "       count++;\n",
    "   }\n",
    "   std::cout << \"Allocating device memory on host..\\n\";\n",
    "   CUDA_CALL(cudaMalloc((void **)&in_d,nBytes));\n",
    "   CUDA_CALL(cudaMalloc((void **)&out_d,nBytes));\n",
    "\n",
    "   //Set up blocks\n",
    "   dim3 dimBlock(block_size,block_size,1);\n",
    "   dim3 dimGrid(sz_x/block_size,sz_y/block_size,1);\n",
    "\n",
    "   std::cout << \"Doing GPU Transpose\\n\";\n",
    "   CUDA_CALL(cudaMemcpy(in_d,m_h,nBytes,cudaMemcpyHostToDevice));\n",
    "    \n",
    "   auto start_d = std::chrono::high_resolution_clock::now();\n",
    "   \n",
    "    /**********************\n",
    "   //transpose<<<dimGrid,dimBlock>>>(in_d,out_d,sz_y,sz_x);\n",
    "   tiled_transpose<<<dimGrid,dimBlock>>>(in_d,out_d,sz_y,sz_x);\n",
    "\n",
    "   cudaError_t err = cudaGetLastError();\n",
    "   if (err != cudaSuccess) {\n",
    "        fprintf(stderr, \"\\n\\nError: %s\\n\\n\", cudaGetErrorString(err)); fflush(stderr); exit(err);   \n",
    "   } \n",
    "   CUDA_CALL(cudaMemcpy(m_h,out_d,nBytes,cudaMemcpyDeviceToHost));\n",
    "   ************************/\n",
    "    \n",
    "   /**********************/\n",
    "   transpose_symmetric<<<dimGrid,dimBlock>>>(in_d,out_d,sz_y,sz_x);    \n",
    "   cudaError_t err = cudaGetLastError();\n",
    "   if (err != cudaSuccess) {\n",
    "        fprintf(stderr, \"\\n\\nError: %s\\n\\n\", cudaGetErrorString(err)); fflush(stderr); exit(err);   \n",
    "   } \n",
    "   CUDA_CALL(cudaMemcpy(m_h,in_d,nBytes,cudaMemcpyDeviceToHost));\n",
    "   /************************/\n",
    "    \n",
    "   auto end_d = std::chrono::high_resolution_clock::now();\n",
    "\n",
    "   std::cout << \"Doing CPU Transpose\\n\";\n",
    "   auto start_h = std::chrono::high_resolution_clock::now();\n",
    "   for (int y=0; y < sz_y; y++){\n",
    "        for (int x=y; x < sz_x; x++){\n",
    "           double temp = m_h[x+sz_x*y];\n",
    "           //std::cout << temp << \" \";\n",
    "           m_h[x+sz_x*y] = m_h[y+sz_y*x];\n",
    "           m_h[y+sz_y*x] = temp;\n",
    "       }\n",
    "       //std::cout << \"\\n\";\n",
    "   }\n",
    "   auto end_h = std::chrono::high_resolution_clock::now();\n",
    "\n",
    "\n",
    "   //Checking errors (should be same values as start)\n",
    "   count = 0;\n",
    "   int errors = 0;\n",
    "   for (int i=0; i < sz_x*sz_y; i++){\n",
    "       if (m_h[i] != count)\n",
    "           errors++;\n",
    "       count++;\n",
    "   }\n",
    "   std::cout << errors << \" Errors found in transpose\\n\";\n",
    "\n",
    "    //Print Timing\n",
    "   std::chrono::duration<double> time_d = end_d - start_d;\n",
    "   std::cout << \"Device time: \" << time_d.count() << \" s\\n\";\n",
    "   std::chrono::duration<double> time_h = end_h - start_h;\n",
    "   std::cout << \"Host time: \" << time_h.count() << \" s\\n\";\n",
    "\n",
    "   cudaFree(in_d);\n",
    "   cudaFree(out_d);\n",
    "   return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "#Compile Cuda\n",
    "!nvcc -std=c++11 -o tiled_transpose NCode/tiled_transpose.cu"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "#Run Example\n",
    "!./tiled_transpose "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "\n",
    "---\n",
    "<a name=1D-wave-Cuda-Code-Optimization></a>\n",
    "# 3. 1D wave Cuda Code Optimization\n",
    "\n",
    "\n",
    "As a group, lets see if we can optimize the code code from lasttime."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%writefile wave_cuda.cu\n",
    "\n",
    "#include <stdio.h>\n",
    "#include <stdlib.h>\n",
    "#include <math.h>\n",
    "#include <cuda.h>\n",
    "#define CUDA_CALL(x) {cudaError_t cuda_error__ = (x); if (cuda_error__) printf(\"CUDA error: \" #x \" returned \\\"%s\\\"\\n\", cudaGetErrorString(cuda_error__));}\n",
    "\n",
    "\n",
    "__global__ void accel_update(double* d_dvdt, double* d_y, int nx, double dx2inv)\n",
    "{\n",
    "        int i = blockDim.x * blockIdx.x + threadIdx.x;\n",
    "        if (i > 0 && i < nx-1)\n",
    "            d_dvdt[i]=(d_y[i+1]+d_y[i-1]-2.0*d_y[i])*(dx2inv);\n",
    "        else\n",
    "            d_dvdt[i] = 0;\n",
    "}\n",
    "\n",
    "__global__ void pos_update(double * d_dvdt, double * d_y, double * d_v, double dt)\n",
    "{\n",
    "        int i = blockDim.x * blockIdx.x + threadIdx.x;\n",
    "        d_v[i] = d_v[i] + dt*d_dvdt[i];\n",
    "        d_y[i]  = d_y[i] + dt*d_v[i];\n",
    "}\n",
    "\n",
    "\n",
    "int main(int argc, char ** argv) {\n",
    "    int nx = 5000;\n",
    "    int nt = 1000000;\n",
    "    int i,it;\n",
    "    double x[nx];\n",
    "    double y[nx];\n",
    "    double v[nx];\n",
    "    double dvdt[nx];\n",
    "    double dt;\n",
    "    double dx;\n",
    "    double max,min;\n",
    "    double dx2inv;\n",
    "    double tmax;\n",
    "\n",
    "    double *d_x, *d_y, *d_v, *d_dvdt;\n",
    "    CUDA_CALL(cudaMalloc((void **)&d_x,nx*sizeof(double)));\n",
    "    CUDA_CALL(cudaMalloc((void **)&d_y,nx*sizeof(double)));\n",
    "    CUDA_CALL(cudaMalloc((void **)&d_v,nx*sizeof(double)));\n",
    "    CUDA_CALL(cudaMalloc((void **)&d_dvdt,nx*sizeof(double)));\n",
    "\n",
    "    max=10.0;\n",
    "    min=0.0;\n",
    "    dx = (max-min)/(double)(nx-1);\n",
    "    x[0] = min;\n",
    "    for(i=1;i<nx-1;i++) {\n",
    "        x[i] = min+(double)i*dx;\n",
    "    }\n",
    "\n",
    "    x[nx-1] = max;\n",
    "    tmax=10.0;\n",
    "    dt= (tmax-0.0)/(double)(nt-1);\n",
    "\n",
    "    for (i=0;i<nx;i++)  {\n",
    "        y[i] = exp(-(x[i]-5.0)*(x[i]-5.0));\n",
    "        v[i] = 0.0;\n",
    "        dvdt[i] = 0.0;\n",
    "    }\n",
    "\n",
    "   CUDA_CALL(cudaMemcpy(d_x,x,nx*sizeof(double),cudaMemcpyHostToDevice));\n",
    "   CUDA_CALL(cudaMemcpy(d_y,y,nx*sizeof(double),cudaMemcpyHostToDevice));\n",
    "   CUDA_CALL(cudaMemcpy(d_v,v,nx*sizeof(double),cudaMemcpyHostToDevice));\n",
    "   CUDA_CALL(cudaMemcpy(d_dvdt,dvdt,nx*sizeof(double),cudaMemcpyHostToDevice));\n",
    "\n",
    "   dx2inv=1.0/(dx*dx);\n",
    "   int block_size=1024;\n",
    "   int block_no = nx/block_size;\n",
    "   dim3 dimBlock(block_size,1,1);\n",
    "   dim3 dimGrid(block_no,1,1);\n",
    "\n",
    "    for(it=0;it<nt-1;it++) {\n",
    "        accel_update<<<dimGrid, dimBlock>>>(d_dvdt, d_y, nx, dx2inv);\n",
    "        pos_update<<<dimGrid, dimBlock>>>(d_dvdt, d_y, d_v, dt);\n",
    "    }\n",
    "\n",
    "   CUDA_CALL(cudaMemcpy(x,d_x,nx*sizeof(double),cudaMemcpyDeviceToHost));\n",
    "   CUDA_CALL(cudaMemcpy(y,d_y,nx*sizeof(double),cudaMemcpyDeviceToHost));\n",
    "\n",
    "    for(i=nx/2-10; i<nx/2+10; i++) {\n",
    "        printf(\"%g %g\\n\",x[i],y[i]);\n",
    "    }\n",
    "\n",
    "    return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!nvcc -std=c++11 -o wave_cuda wave_cuda.cu"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%time\n",
    "!./wave_cuda"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "\n",
    "-----\n",
    "### Congratulations, we're done!\n",
    "\n",
    "If you attend class in-person then have one of the instructors check your notebook and sign you out before leaving class. If you are attending asynchronously, turn in your assignment using D2L."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Course Resources:\n",
    "\n",
    "\n",
    " - [Website](https://msu-cmse-courses.github.io/cmse802-f20-student/)\n",
    " - [ZOOM](https://msu.zoom.us/j/98207034052)\n",
    " - [JargonJar](https://docs.google.com/document/d/1ahg48CCFhRzUL-QIHzlt_KEf1XqsCasFBU4iePHhcug/edit#)\n",
    " - [GIT](https://gitlab.msu.edu/colbrydi/cmse401-s21.git)\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Written by Dr. Dirk Colbry, Michigan State University\n",
    "<a rel=\"license\" href=\"http://creativecommons.org/licenses/by-nc/4.0/\"><img alt=\"Creative Commons License\" style=\"border-width:0\" src=\"https://i.creativecommons.org/l/by-nc/4.0/88x31.png\" /></a><br />This work is licensed under a <a rel=\"license\" href=\"http://creativecommons.org/licenses/by-nc/4.0/\">Creative Commons Attribution-NonCommercial 4.0 International License</a>."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "----"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
